import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:mobx/mobx.dart';

part 'machine.freezed.dart';

typedef ImmediateCb<Value, Event> = Value? Function();
typedef OnMachineEventCb<Value, Event> = Value? Function(
    MachineEvent<Value, Event> event, AddEventSideEffect sideEffect);
typedef OnEventCb<Value, Event> = Value? Function(Event event);
typedef AddEventSideEffect = void Function(void Function() cb);
typedef EventSideEffect = void Function();
typedef CleanUpLocals = void Function();
typedef OnChangeCb<Value, Event> = StateConfig<Value, Event> Function(
    Value value);

class MachineState<Value, Event> {
  MachineState({
    required this.value,
    required this.config,
    required this.actions,
  });
  final Value value;
  final StateConfig<Value, Event> config;
  final List<StateMachineAction<Value, Event>> actions;
}

abstract class Machine {
  static MachineState<Value, Event> init<Value, Event>(
      OnChangeCb<Value, Event> onChange, Value initial) {
    final config = onChange(initial);
    final immediateTransition = immediate(onChange, initial, config, config);
    return MachineState<Value, Event>(
      value: immediateTransition.value,
      config: immediateTransition.cfg,
      actions: [
        if (immediateTransition.cfg._enter.isNotEmpty)
          ...immediateTransition.cfg._enter,
      ],
    );
  }

  static ImmediateTransition<Value, Event> immediate<Value, Event>(
      OnChangeCb<Value, Event> onChange,
      Value value,
      StateConfig<Value, Event> cur,
      StateConfig<Value, Event> startFrom) {
    if (true == cur._immediate.isNotEmpty) {
      for (var i = 0; i < cur._immediate.length; i++) {
        final next = cur._immediate[i]();
        if (null != next) {
          if (cur is StateConfigWithLocals<Value, Event>) {
            /// cleanup mutable data
            for (var clean in cur._cleanUp) {
              clean();
            }
          }

          var nextConfig = onChange(next);

          return immediate(onChange, next, nextConfig, startFrom);
        }
      }
    }

    return ImmediateTransition<Value, Event>(
      cfg: cur,
      value: value,
    );
  }

  @protected
  @visibleForTesting
  static MachineState<Value, Event> transition<Value, Event>(
    OnChangeCb<Value, Event> onChange,
    MachineState<Value, Event> state,
    MachineEvent<Value, Event> event,
  ) {
    List<EventSideEffect> sideEffects = [];
    final maybeValue = event.map(
      send: (obj) => state.config._getValueFromEvent(event, sideEffects),
      iToValue: (obj) => obj.value,
    );

    if (null == maybeValue) return state;
    final curConfig = state.config;

    if (curConfig is StateConfigWithLocals<Value, Event>) {
      /// cleanup mutable data
      for (var clean in curConfig._cleanUp) {
        clean();
      }
    }

    var nextValue = maybeValue;

    var nextConfig = onChange(nextValue);
    final immediateTransition =
        immediate(onChange, nextValue, nextConfig, nextConfig);

    nextValue = immediateTransition.value;
    nextConfig = immediateTransition.cfg;

    return MachineState<Value, Event>(
      value: nextValue,
      config: nextConfig,
      actions: [
        ...sideEffects.map(
            (cb) => StateMachineActionEventSideEffect<Value, Event>(cb: cb)),
        if (curConfig._leave.isNotEmpty) ...curConfig._leave,
        if (nextConfig._enter.isNotEmpty) ...nextConfig._enter,
      ],
    );
  }

  static MachineState<Value, Event> send<Value, Event>(
    OnChangeCb<Value, Event> onChange,
    MachineState<Value, Event> state,
    Event event,
  ) =>
      transition<Value, Event>(
          onChange, state, MachineEvent<Value, Event>.send(event: event));
}

class StateConfig<Value, Event> {
  final List<ImmediateCb<Value, Event>> _immediate;
  final List<OnMachineEventCb<Value, Event>> _onEvent;
  final List<StateMachineActionEnter<Value, Event>> _enter;
  final List<StateMachineActionLeave<Value, Event>> _leave;

  StateConfig({
    List<ImmediateCb<Value, Event>>? immediate,
    List<OnMachineEventCb<Value, Event>>? onEvent,
    List<StateMachineActionEnter<Value, Event>>? enter,
    List<StateMachineActionLeave<Value, Event>>? leave,
  })  : _immediate = [],
        _onEvent = [],
        _enter = [],
        _leave = [];

  Value? _getValueFromEvent(MachineEvent<Value, Event> machineEvent,
      List<void Function()> sideEffects) {
    Value? ev;

    /// first event not to return null will be taken
    if (true == _onEvent.isNotEmpty) {
      for (var i = 0; i < _onEvent.length; i++) {
        ev = _onEvent[i](machineEvent, sideEffects.add);

        if (null != ev) break;
      }
    }

    return ev;
  }

  void immediate(ImmediateCb<Value, Event> cb) => _immediate.add(cb);

  void onEvent(Value? Function(Event event, AddEventSideEffect sideEffect) cb) {
    _onEvent.add((ev, ef) {
      return ev.maybeMap<Value?>(
        send: (obj) => cb(obj.event, ef),
        orElse: () => null,
      );
    });
  }

  void enter(void Function(IOnEnter<Value, Event> enter) cb) {
    _enter.add(StateMachineActionEnter<Value, Event>(cb: cb));
  }

  void leave(void Function() cb) {
    _leave.add(StateMachineActionLeave<Value, Event>(cb: cb));
  }
}

class StateConfigWithLocals<Value, Event> extends StateConfig<Value, Event> {
  final List<CleanUpLocals> _cleanUp = [];
  void cleanUp(CleanUpLocals cb) => _cleanUp.add(cb);
}

@freezed
class ImmediateTransition<Value, Event>
    with _$ImmediateTransition<Value, Event> {
  factory ImmediateTransition({
    required Value value,
    required StateConfig<Value, Event> cfg,
  }) = _ImmediateTransition<Value, Event>;
}

@freezed
class StateMachineAction<Value, Event> with _$StateMachineAction<Value, Event> {
  factory StateMachineAction.enter({
    required void Function(IOnEnter<Value, Event> enter) cb,
  }) = StateMachineActionEnter<Value, Event>;

  factory StateMachineAction.leave({
    required void Function() cb,
  }) = StateMachineActionLeave<Value, Event>;

  factory StateMachineAction.eventSideEffetct({
    required void Function() cb,
  }) = StateMachineActionEventSideEffect<Value, Event>;
}

@freezed
class MachineEvent<Value, Event> with _$MachineEvent<Value, Event> {
  factory MachineEvent.send({required Event event}) = MESend<Value, Event>;

  /// machine internal event
  factory MachineEvent.iToValue({required Value value}) =
      MEInternalToValue<Value, Event>;
}

abstract class IOnEnter<Value, Event> {
  bool get isActive;
  void send(Event event);
  void next(Value stateValue);
  void cleanUp(void Function() cb);
}

class SOnEnter<Value, Event> extends IOnEnter<Value, Event> {
  SOnEnter(this._service) {
    _stateOnEnter = _service.state$.value;
  }

  final Service<Value, Event> _service;
  final List<void Function()> _cleanUp = [];
  late final MachineState<Value, Event> _stateOnEnter;
  bool _disposed = false;

  void dispose() {
    _disposed = true;
    for (var d in _cleanUp) {
      d();
    }
    _cleanUp.clear();
  }

  @override
  bool get isActive => !_disposed && _stateOnEnter == _service.state$.value;

  @override
  void next(Value stateValue) {
    if (!isActive) {
      throw StateError('''
It is not possible to transition to another State using .next() after State of "${_stateOnEnter.value.runtimeType}" has been left.
This happened in Service "${_service.runtimeType}".
Consider using enter.isActive before.''');
    }
    _service.sendMachineEvent(
        MachineEvent<Value, Event>.iToValue(value: stateValue));
  }

  @override
  void cleanUp(void Function() cb) {
    if (!isActive) {
      throw StateError('''
It is not possible to add a cleanUp callback after State of "${_stateOnEnter.value.runtimeType}" has been left.
This happened in Service "${_service.runtimeType}".
Consider using enter.isActive before.''');
    }

    _cleanUp.add(cb);
  }

  @override
  void send(Event event) {
    if (!isActive) {
      throw StateError('''
It is not possible to .send($event) after State of "${_stateOnEnter.value.runtimeType}" has been left.
This happened in Service "${_service.runtimeType}".
Consider using enter.isActive before.''');
    }
    _service.send(event);
  }
}

class _CreateConfig<Value, Event> {
  StateConfig<Value, Event> create() => StateConfig<Value, Event>();
  StateConfig<Value, Event> withLocals(
      void Function(StateConfigWithLocals<Value, Event> config) cb) {
    final cfg = StateConfigWithLocals<Value, Event>();
    cb(cfg);
    return cfg;
  }
}

abstract class Service<Value, Event> {
  Service() {
    final state = Machine.init<Value, Event>(onChange, initialValue);
    state$ = Observable<MachineState<Value, Event>>(
      state,
      context: _ctx,
      equals: (_, __) => false,
      name: '$runtimeType.Observable.state',
    );

    executeActions(state);
  }

  Value get initialValue;

  StateConfig<Value, Event> onChange(Value value);

  static final _ctx = ReactiveContext(config: ReactiveConfig.main);

  /// Helper function to easily create a [StateConfig] or [StateConfigWithLocals] object.
  @protected
  late final config = _CreateConfig<Value, Event>();

  @protected
  late final ActionController ctrl =
      ActionController(context: _ctx, name: '$runtimeType.ActionController');

  @protected
  @visibleForTesting
  late final Observable<MachineState<Value, Event>> state$;

  final List<SOnEnter<Value, Event>> _activeEnter = [];
  final List<ReactionDisposer> _watchDisposer = [];

  bool get hasListener => _watchDisposer.isNotEmpty;

  bool _mounted = true;
  bool get mounted => _mounted;

  @mustCallSuper
  void dispose() {
    for (var enter in _activeEnter) {
      enter.dispose();
    }
    _activeEnter.clear();
    _watchDisposer
      ..forEach((cb) => cb())
      ..clear();
    _mounted = false;
  }

  @protected
  void toState(MachineState<Value, Event> state,
      List<StateMachineAction<Value, Event>> actions) {
    final info = ctrl.startAction(name: '$runtimeType.toState()');
    try {
      for (var enter in _activeEnter) {
        enter.dispose();
      }
      _activeEnter.clear();

      state$.value = state;
    } finally {
      ctrl.endAction(info);
    }

    executeActions(state);
  }

  @protected
  void executeActions(MachineState<Value, Event> state) {
    for (var a in state.actions) {
      if (state$.value != state) break;
      a.map(
        enter: (obj) {
          final enter = SOnEnter<Value, Event>(this);
          _activeEnter.add(enter);
          obj.cb(enter);
        },
        eventSideEffetct: (obj) => obj.cb(),
        leave: (obj) => obj.cb(),
      );
    }
  }

  @protected
  @visibleForTesting
  void sendMachineEvent(MachineEvent<Value, Event> event) {
    final nextState =
        Machine.transition<Value, Event>(onChange, state$.value, event);
    if (nextState == state$.value) return;
    toState(nextState, nextState.actions);
  }

  void send(Event event) {
    final nextState = Machine.send<Value, Event>(onChange, state$.value, event);
    if (nextState == state$.value) return;
    toState(nextState, nextState.actions);
  }

  void Function() watch(
    void Function(MachineState<Value, Event> value) cb, {
    bool fireImmediately = true,
  }) {
    final disposer = reaction<MachineState<Value, Event>>(
      (_) => state$.value,
      cb,
      context: _ctx,
      equals: (_, __) => false,
      fireImmediately: fireImmediately,
      name: '$runtimeType.watch',
    );

    _watchDisposer.add(disposer);

    return () {
      if (disposer.reaction.isDisposed) return;
      _watchDisposer.removeWhere((element) => element == disposer);
      disposer();
    };
  }
}
