import 'package:endlich/src/machine.dart';
import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:test/test.dart';

import 'machine_test.mocks.dart';

part 'machine_test.freezed.dart';

@GenerateMocks([Fn, OnEvent, SimpleWatcher, SimpleOnEnterCb])
void main() {
  group('StateMachine', () {
    test('initial state', () {
      final state = _init(_onChange(), SV.foo());
      expect(state.value, SV.foo());
    });

    test('MachineEvent To Value is always possible', () {
      final onChange = _onChange();
      final state = _sendMachineEvent(onChange, _init(onChange, SV.foo()),
          MachineEvent.iToValue(value: SV.bar()));
      expect(state.value, SV.bar());
    });

    group('send event', () {
      test('will result in transition', () {
        final onChange = _onChange(
          onFoo: (foo, cfg) => cfg
            ..onEvent(
              (event, _) => event.maybeMap(
                orElse: () => null,
                next: (_) => SV.baz(),
                toBar: (_) => SV.bar(),
              ),
            ),
        );

        expect(
            _send(onChange, _init(onChange), TestEvent.next()).value, SV.baz());
        expect(_send(onChange, _init(onChange), TestEvent.toBar()).value,
            SV.bar());
        expect(_send(onChange, _init(onChange), TestEvent.toBaz()).value,
            SV.foo());
      });

      test('will transition on first event to not return null', () {
        final ev1 = MockOnEvent();
        final ev2 = MockOnEvent();
        when(ev1(any, any)).thenAnswer((realInvocation) => null);
        when(ev2(any, any)).thenAnswer((realInvocation) => null);
        final onChange = _onChange(
            onFoo: (foo, cfg) => cfg
              ..onEvent(ev1)
              ..onEvent(ev2)
              ..onEvent((event, _) => SV.bar()));

        expect(
            _send(onChange, _init(onChange), TestEvent.next()).value, SV.bar());

        verify(ev1(any, any)).called(1);
        verify(ev2(any, any)).called(1);
      });
    });

    group('immediate transition', () {
      test('on initial value', () {
        final onChange = _onChange(
          onFoo: (obj, cfg) => cfg..immediate(() => SV.bar()),
          onBar: (obj, cfg) => cfg..immediate(() => SV.baz()),
        );

        expect(_init(onChange).value, SV.baz());
      });

      test('on transition', () {
        final onChange = _onChange(
          onBar: (obj, cfg) => cfg..immediate(() => SV.baz()),
        );

        expect(
            _sendMachineEvent(onChange, _init(onChange),
                    MachineEvent.iToValue(value: SV.bar()))
                .value,
            SV.baz());
      });

      test('does not transition on null', () {
        final onChange = _onChange(
          onFoo: (obj, cfg) => cfg..immediate(() => null),
        );
        expect(
          _init(onChange).value,
          SV.foo(),
        );
      });
    });
  });

  group('StateConfigWithLocals cleanUp', () {
    OnChangeCb<SV, TestEvent> $onChangeWithLocals({
      StateConfig<SV, TestEvent> Function(
              _Foo obj, StateConfigWithLocals<SV, TestEvent> cfg)?
          onFoo,
      StateConfig<SV, TestEvent> Function(
              _Bar obj, StateConfigWithLocals<SV, TestEvent> cfg)?
          onBar,
      StateConfig<SV, TestEvent> Function(
              _Baz obj, StateConfigWithLocals<SV, TestEvent> cfg)?
          onBaz,
    }) =>
        (s) => s.map(
              foo: (obj) {
                final cfg = StateConfigWithLocals<SV, TestEvent>();
                return onFoo?.call(obj, cfg) ?? cfg;
              },
              bar: (obj) {
                final cfg = StateConfigWithLocals<SV, TestEvent>();
                return onBar?.call(obj, cfg) ?? cfg;
              },
              baz: (obj) {
                final cfg = StateConfigWithLocals<SV, TestEvent>();
                return onBaz?.call(obj, cfg) ?? cfg;
              },
            );
    test('while enter initial state and transition', () {
      final fooClean = MockFn();
      final barClean = MockFn();
      final bazClean = MockFn();
      final onChange = $onChangeWithLocals(
        onFoo: (foo, cfg) => cfg..cleanUp(fooClean),
        onBar: (bar, cfg) => cfg..cleanUp(barClean),
        onBaz: (baz, cfg) => cfg..cleanUp(bazClean),
      );
      var s = _init(onChange);

      verifyZeroInteractions(fooClean);
      verifyZeroInteractions(barClean);
      verifyZeroInteractions(bazClean);

      s = _sendMachineEvent(
          onChange, s, MachineEvent.iToValue(value: SV.bar()));
      verify(fooClean()).called(1);
      verifyZeroInteractions(barClean);
      verifyZeroInteractions(bazClean);

      verifyNoMoreInteractions(fooClean);

      s = _sendMachineEvent(
          onChange, s, MachineEvent.iToValue(value: SV.bar()));
      verify(barClean()).called(1);
      verifyZeroInteractions(bazClean);

      s = _sendMachineEvent(
          onChange, s, MachineEvent.iToValue(value: SV.baz()));
      verify(barClean()).called(1);
      verifyZeroInteractions(bazClean);
    });

    test('on immediate transition during machine creation', () {
      final fooClean = MockFn();
      final barClean = MockFn();
      final bazClean = MockFn();
      var toSelf = false;
      final onChange = $onChangeWithLocals(
        onFoo: (foo, cfg) => cfg
          ..cleanUp(fooClean)
          ..immediate(() => SV.bar()),
        onBar: (bar, cfg) => cfg
          ..cleanUp(barClean)
          ..immediate(() {
            if (!toSelf) {
              toSelf = true;
              return SV.bar();
            }
            return SV.baz();
          }),
        onBaz: (baz, cfg) => cfg..cleanUp(bazClean),
      );
      _init(onChange);

      verifyInOrder([
        fooClean(),
        barClean(),
        barClean(),
      ]);
      verifyZeroInteractions(bazClean);
    });

    test('on immediate transition on event', () {
      final fooClean = MockFn();
      final barClean = MockFn();
      final bazClean = MockFn();
      var toSelf = false;
      final onChange = $onChangeWithLocals(
        onFoo: (foo, cfg) => cfg..cleanUp(fooClean),
        onBar: (bar, cfg) => cfg
          ..cleanUp(barClean)
          ..immediate(() {
            if (!toSelf) {
              toSelf = true;
              return SV.bar();
            }
            return SV.baz();
          }),
        onBaz: (baz, cfg) => cfg..cleanUp(bazClean),
      );

      _sendMachineEvent(
          onChange, _init(onChange), MachineEvent.iToValue(value: SV.bar()));

      verifyInOrder([
        fooClean(),
        barClean(),
        barClean(),
      ]);
      verifyZeroInteractions(bazClean);
    });
  });

  group('Current.actions', () {
    test('after creating machine', () {
      enter(IOnEnter<SV, TestEvent> e) {}
      leave() {}
      final onChange = _onChange(
        onFoo: (foo, cfg) => cfg
          ..enter(enter)
          ..leave(leave),
      );

      expect(_init(onChange).actions,
          equals([isA<_AEnter>().having((s) => s.cb, 'cb', enter)]));
    });

    test('after creating machine with immediate transition', () {
      enterFoo(IOnEnter<SV, TestEvent> e) {}
      leaveFoo() {}
      enterBar(IOnEnter<SV, TestEvent> e) {}
      leaveBar() {}
      final onChange = _onChange(
        onFoo: (foo, cfg) => cfg
          ..immediate(() => SV.bar())
          ..enter(enterFoo)
          ..leave(leaveFoo),
        onBar: (bar, cfg) => cfg
          ..enter(enterBar)
          ..leave(leaveBar),
      );
      expect(_init(onChange).actions,
          equals([isA<_AEnter>().having((s) => s.cb, 'cb', enterBar)]));
    });

    test('after transition', () {
      enterFoo(IOnEnter<SV, TestEvent> e) {}
      leaveFoo() {}
      sideEffectFoo() {}
      enterBar(IOnEnter<SV, TestEvent> e) {}
      leaveBar() {}
      final onChange = _onChange(
        onFoo: (foo, cfg) => cfg
          ..onEvent((event, sideEffect) {
            sideEffect(sideEffectFoo);
            return SV.bar();
          })
          ..enter(enterFoo)
          ..leave(leaveFoo),
        onBar: (bar, cfg) => cfg
          ..enter(enterBar)
          ..leave(leaveBar),
      );
      var s = _send(onChange, _init(onChange), TestEvent.next());

      expect(
          s.actions,
          equals([
            isA<_AEventSE>().having((s) => s.cb, 'cb', sideEffectFoo),
            isA<_ALeave>().having((s) => s.cb, 'cb', leaveFoo),
            isA<_AEnter>().having((s) => s.cb, 'cb', enterBar),
          ]));
    });

    test('after transition and immediate', () {
      enterFoo(IOnEnter<SV, TestEvent> e) {}
      leaveFoo() {}
      sideEffectFoo() {}
      enterBar(IOnEnter<SV, TestEvent> e) {}
      leaveBar() {}
      enterBaz(IOnEnter<SV, TestEvent> e) {}
      leaveBaz() {}
      final onChange = _onChange(
        onFoo: (foo, cfg) => cfg
          ..onEvent((event, sideEffect) {
            sideEffect(sideEffectFoo);
            return SV.bar();
          })
          ..enter(enterFoo)
          ..leave(leaveFoo),
        onBar: (bar, cfg) => cfg
          ..immediate(() => SV.baz())
          ..enter(enterBar)
          ..leave(leaveBar),
        onBaz: (baz, cfg) => cfg
          ..enter(enterBaz)
          ..leave(leaveBaz),
      );

      var s = _send(onChange, _init(onChange), TestEvent.next());

      expect(
          s.actions,
          equals([
            isA<_AEventSE>().having((s) => s.cb, 'cb', sideEffectFoo),
            isA<_ALeave>().having((s) => s.cb, 'cb', leaveFoo),
            isA<_AEnter>().having((s) => s.cb, 'cb', enterBaz)
          ]));
    });
  });

  group('Service', () {
    test('initialize', () {
      final s = TestService();
      expect(
          s.state$.value,
          isA<MachineState<SV, TestEvent>>()
              .having((s) => s.value, 'value', SV.foo()));
      expect(s.hasListener, isFalse);
    });

    group('transition', () {
      test('will execute actions', () {
        final fooEnter = MockSimpleOnEnterCb();
        final fooSE = MockFn();
        final fooLeave = MockFn();
        final barEnter = MockSimpleOnEnterCb();
        final s = TestService(
          onFoo: (foo, cfg) => cfg
            ..onEvent((event, sideEffect) {
              sideEffect(fooSE);
              return SV.bar();
            })
            ..leave(fooLeave)
            ..enter(fooEnter),
          onBar: (bar, cfg) => cfg..enter(barEnter),
        );

        s.send(TestEvent.next());
        verifyInOrder([
          fooEnter(any),
          fooSE(),
          fooLeave(),
          barEnter(any),
        ]);
      });
    });

    group('sending event', () {
      test('will transition', () {
        final s = TestService(
          onFoo: (foo, cfg) => cfg..onEvent((event, sideEffect) => SV.bar()),
        );
        s.send(TestEvent.next());
        expect(
            s.state$.value,
            isA<MachineState<SV, TestEvent>>()
                .having((s) => s.value, 'value', SV.bar()));
      });

      test('from machine event', () {
        final s = TestService();

        s.sendMachineEvent(MachineEvent.iToValue(value: SV.bar()));
        expect(
            s.state$.value,
            isA<MachineState<SV, TestEvent>>()
                .having((s) => s.value, 'value', SV.bar()));
      });

      test(
          'synchronously in an action will prevent further execution of other following enter cb',
          () {
        final e1 = MockSimpleOnEnterCb();
        final e2 = MockSimpleOnEnterCb();
        TestService(
          onFoo: (state, config) => config
            ..enter(e1)
            ..enter((enter) {
              enter.next(SV.bar());
            })
            ..enter(e2),
        );
        verify(e1(any)).called(1);
        verifyNever(e2(any));
      });
    });

    group('.watch()', () {
      test('automatically receive current value on watch', () {
        final watcher = MockSimpleWatcher();
        final s = TestService();
        s.watch(watcher);

        verify(watcher(argThat(isA<MachineState<SV, TestEvent>>()
                .having((s) => s.value, 'value', SV.foo()))))
            .called(1);
      });

      test('skip receiving value on watch', () {
        final watcher = MockSimpleWatcher();
        final s = TestService();
        s.watch(watcher, fireImmediately: false);
        verifyNever(watcher(argThat(isA<MachineState<SV, TestEvent>>()
            .having((s) => s.value, 'value', SV.foo()))));
      });

      test('can watch updates', () {
        final watcher = MockSimpleWatcher();
        final s = TestService(
          onFoo: (state, config) => config..onEvent((event, _) => SV.bar()),
          onBar: (state, config) => config..onEvent((event, _) => SV.baz()),
          onBaz: (state, config) => config..onEvent((event, _) => SV.foo()),
        );
        final d = s.watch(watcher);
        s..send(TestEvent.next())..send(TestEvent.next());
        d();
        s.send(TestEvent.next());
        verifyInOrder([
          watcher(argThat(isA<MachineState<SV, TestEvent>>()
              .having((s) => s.value, 'value', SV.foo()))),
          watcher(argThat(isA<MachineState<SV, TestEvent>>()
              .having((s) => s.value, 'value', SV.bar()))),
          watcher(argThat(isA<MachineState<SV, TestEvent>>()
              .having((s) => s.value, 'value', SV.baz()))),
        ]);
      });

      test('not getting called on immediate transition', () {
        final watcher = MockSimpleWatcher();
        final s = TestService(
          onFoo: (state, config) => config..onEvent((event, _) => SV.bar()),
          onBar: (state, config) => config..immediate(() => SV.baz()),
          onBaz: (state, config) => config..immediate(() => SV.foo()),
        );
        s.watch(watcher);
        s.send(TestEvent.next());
        verifyNever(watcher(argThat(isA<MachineState<SV, TestEvent>>()
            .having((s) => s.value, 'value', SV.bar()))));
        verifyNever(watcher(argThat(isA<MachineState<SV, TestEvent>>()
            .having((s) => s.value, 'value', SV.baz()))));
        verifyInOrder([
          watcher(argThat(isA<MachineState<SV, TestEvent>>()
              .having((s) => s.value, 'value', SV.foo()))),
          watcher(argThat(isA<MachineState<SV, TestEvent>>()
              .having((s) => s.value, 'value', SV.foo()))),
        ]);
      });

      test('.hasListener', () {
        final s = TestService();
        expect(s.hasListener, isFalse);

        final sub1 = s.watch((state) {});
        expect(s.hasListener, isTrue);

        final sub2 = s.watch((state) {});
        expect(s.hasListener, isTrue);

        sub2();
        expect(s.hasListener, isTrue);

        sub1();
        expect(s.hasListener, isFalse);

        s.watch((state) {});
        expect(s.hasListener, isTrue);

        s.dispose();
        expect(s.hasListener, isFalse);
      });

      test('dispose machine and remove listener after will not throw', () {
        final s = TestService();
        final disp = s.watch((state) {});
        s.dispose();
        expect(() => disp(), isNot(throwsException));
        expect(s.hasListener, isFalse);
      });

      group('enter', () {
        test('can transition using .send()', () {
          final watcher = MockSimpleWatcher();
          final s = TestService(
            onFoo: (state, config) => config
              ..onEvent(
                  (event, sideEffect) => event.mapOrNull(next: (_) => SV.baz()))
              ..enter((enter) => enter.send(TestEvent.next())),
          );
          s.watch(watcher);
          verify(watcher(argThat(isA<MachineState<SV, TestEvent>>()
                  .having((s) => s.value, 'value', SV.baz()))))
              .called(1);
        });

        test('using .send() will throw an error after state has been left', () {
          late IOnEnter<SV, TestEvent> fooOnEnter;
          final s = TestService(
            onFoo: (state, config) => config
              ..onEvent(
                (event, sideEffect) => event.mapOrNull(
                  next: (_) => SV.baz(),
                  toBar: (_) => SV.bar(),
                ),
              )
              ..enter((enter) {
                fooOnEnter = enter;
              }),
          );

          s.send(TestEvent.next());

          expect(() => fooOnEnter.send(TestEvent.toBar()), throwsStateError);
        });

        test('can transition using .next()', () {
          final watcher = MockSimpleWatcher();
          final s = TestService(
            onFoo: (state, config) => config
              ..enter((enter) {
                enter.next(SV.bar());
              }),
          );
          s.watch(watcher);

          verifyNever(watcher(argThat(isA<MachineState<SV, TestEvent>>()
              .having((s) => s.value, 'value', SV.foo()))));
          verifyInOrder([
            watcher(argThat(isA<MachineState<SV, TestEvent>>()
                .having((s) => s.value, 'value', SV.bar()))),
          ]);
        });

        test('using .next() will throw an error after state has been left', () {
          late IOnEnter<SV, TestEvent> fooOnEnter;
          final s = TestService(
            onFoo: (state, config) => config
              ..onEvent(
                (event, sideEffect) => event.mapOrNull(
                  next: (_) => SV.baz(),
                  toBar: (_) => SV.bar(),
                ),
              )
              ..enter((enter) {
                fooOnEnter = enter;
              }),
          );

          s.send(TestEvent.next());

          expect(() => fooOnEnter.next(SV.bar()), throwsStateError);
        });

        test('.isActive', () {
          late IOnEnter<SV, TestEvent> fooOnEnter;
          late IOnEnter<SV, TestEvent> barOnEnter;
          final s = TestService(
            onFoo: (state, config) => config
              ..onEvent((event, sideEffect) => SV.bar())
              ..enter((enter) {
                fooOnEnter = enter;
              }),
            onBar: (state, config) => config
              ..onEvent((event, sideEffect) => SV.baz())
              ..enter((enter) {
                barOnEnter = enter;
              }),
          );
          expect(fooOnEnter.isActive, isTrue);

          s.send(TestEvent.next());
          expect(fooOnEnter.isActive, isFalse);
          expect(barOnEnter.isActive, isTrue);
        });

        test('.isActive on service dispose', () {
          late IOnEnter<SV, TestEvent> fooOnEnter;
          final s = TestService(
            onFoo: (state, config) => config
              ..onEvent((event, sideEffect) => SV.bar())
              ..enter((enter) {
                fooOnEnter = enter;
              }),
          );
          s.dispose();
          expect(fooOnEnter.isActive, isFalse);
        });

        test('.cleanUp()', () {
          final fooCleanUp = MockFn();
          final s = TestService(
            onFoo: (state, config) => config
              ..onEvent((event, sideEffect) => SV.bar())
              ..enter((enter) {
                enter.cleanUp(fooCleanUp);
              }),
          );

          s.send(TestEvent.next());
          verify(fooCleanUp()).called(1);
        });

        test('using .cleanUp() throws error after state has been left', () {
          late IOnEnter<SV, TestEvent> fooOnEnter;
          final s = TestService(
            onFoo: (state, config) => config
              ..onEvent(
                (event, sideEffect) => event.mapOrNull(
                  next: (_) => SV.baz(),
                ),
              )
              ..enter((enter) {
                fooOnEnter = enter;
              }),
          );

          s.send(TestEvent.next());
          expect(() => fooOnEnter.cleanUp(() {}), throwsStateError);
        });

        test('.cleanUp() on service dispose', () {
          final fooCleanUp = MockFn();
          final s = TestService(
            onFoo: (state, config) => config
              ..onEvent((event, sideEffect) => SV.bar())
              ..enter((enter) {
                enter.cleanUp(fooCleanUp);
              }),
          );
          s.dispose();
          verify(fooCleanUp()).called(1);
        });
      });
    });
  });
}

OnChangeCb<SV, TestEvent> _onChange({
  StateConfig<SV, TestEvent> Function(_Foo obj, StateConfig<SV, TestEvent> cfg)?
      onFoo,
  StateConfig<SV, TestEvent> Function(_Bar obj, StateConfig<SV, TestEvent> cfg)?
      onBar,
  StateConfig<SV, TestEvent> Function(_Baz obj, StateConfig<SV, TestEvent> cfg)?
      onBaz,
}) =>
    (s) => s.map(
          foo: (obj) =>
              onFoo?.call(obj, StateConfig<SV, TestEvent>()) ??
              StateConfig<SV, TestEvent>(),
          bar: (obj) =>
              onBar?.call(obj, StateConfig<SV, TestEvent>()) ??
              StateConfig<SV, TestEvent>(),
          baz: (obj) =>
              onBaz?.call(obj, StateConfig<SV, TestEvent>()) ??
              StateConfig<SV, TestEvent>(),
        );

MachineState<SV, TestEvent> _init(OnChangeCb<SV, TestEvent> onChange,
    [SV? initial]) {
  return Machine.init<SV, TestEvent>(onChange, initial ?? SV.foo());
}

MachineState<SV, TestEvent> _sendMachineEvent(
    OnChangeCb<SV, TestEvent> onChange,
    MachineState<SV, TestEvent> state,
    MachineEvent<SV, TestEvent> event) {
  return Machine.transition<SV, TestEvent>(onChange, state, event);
}

MachineState<SV, TestEvent> _send(OnChangeCb<SV, TestEvent> onChange,
    MachineState<SV, TestEvent> state, TestEvent event) {
  return Machine.send<SV, TestEvent>(onChange, state, event);
}

class TestService extends Service<SV, TestEvent> {
  TestService({
    this.onFoo,
    this.onBar,
    this.onBaz,
  }) : super();

  @override
  SV get initialValue => SV.foo();

  StateConfig<SV, TestEvent> Function(_Foo obj, StateConfig<SV, TestEvent> cfg)?
      onFoo;
  StateConfig<SV, TestEvent> Function(_Bar obj, StateConfig<SV, TestEvent> cfg)?
      onBar;
  StateConfig<SV, TestEvent> Function(_Baz obj, StateConfig<SV, TestEvent> cfg)?
      onBaz;

  @override
  StateConfig<SV, TestEvent> onChange(SV value) => value.map(
        foo: (obj) =>
            onFoo?.call(obj, StateConfig<SV, TestEvent>()) ??
            StateConfig<SV, TestEvent>(),
        bar: (obj) =>
            onBar?.call(obj, StateConfig<SV, TestEvent>()) ??
            StateConfig<SV, TestEvent>(),
        baz: (obj) =>
            onBaz?.call(obj, StateConfig<SV, TestEvent>()) ??
            StateConfig<SV, TestEvent>(),
      );
}

@freezed
class SV with _$SV {
  factory SV.foo() = _Foo;
  factory SV.bar() = _Bar;
  factory SV.baz() = _Baz;
}

@freezed
class TestEvent with _$TestEvent {
  factory TestEvent.next() = _Next;
  factory TestEvent.toFoo() = _ToFoo;
  factory TestEvent.toBar() = _ToBar;
  factory TestEvent.toBaz() = _ToBaz;
}

typedef _AEnter = StateMachineActionEnter<SV, TestEvent>;
typedef _ALeave = StateMachineActionLeave<SV, TestEvent>;
typedef _AEventSE = StateMachineActionEventSideEffect<SV, TestEvent>;

class Fn extends Mock {
  void call();
}

class OnEvent extends Mock {
  SV? call(TestEvent event, AddEventSideEffect sideEffect);
}

class SimpleWatcher extends Mock {
  void call(MachineState<SV, TestEvent> state);
}

class SimpleOnEnterCb extends Mock {
  void call(IOnEnter<SV, TestEvent> enter);
}
